"""
Link layer in UPDI protocol stack
"""
from logging import getLogger

from pymcuprog.pymcuprog_errors import PymcuprogSerialUpdiError, PymcuprogSerialUpdiProtocolError
from . import constants


class UpdiDatalink:
    """
    UPDI data link class handles the UPDI data protocol within the device
    """

    LDCS_RESPONSE_BYTES = 1

    def __init__(self):
        self.logger = getLogger(__name__)
        self.updi_phy = None

    def set_physical(self, physical):
        """
        Inject a serial-port based physical layer for use by this DL

        :param physical: physical object to use
        :type physical: UpdiPhysical
        """
        self.updi_phy = physical

    def _init_session_parameters(self):
        """
        Set the inter-byte delay bit and disable collision detection
        """
        self.stcs(constants.UPDI_CS_CTRLB, 1 << constants.UPDI_CTRLB_CCDETDIS_BIT)
        self._enable_ack()

    def _disable_ack(self):
        """
        Disables ACKs on write to reduce latency for writing blocks
        TODO: could read-modify-write
        """
        self.stcs(constants.UPDI_CS_CTRLA, 1 << constants.UPDI_CTRLA_IBDLY_BIT |
            1 << constants.UPDI_CTRLA_RSD_BIT)

    def _enable_ack(self):
        """
        Enables ACKs on write by default
        TODO: could read-modify-write
        """
        self.stcs(constants.UPDI_CS_CTRLA, 1 << constants.UPDI_CTRLA_IBDLY_BIT)

    def init_datalink(self):
        """
        Init DL layer

        :raises: PymcuprogSerialUpdiError if initialization fails
        """
        self._init_session_parameters()
        # Check
        if not self._check_datalink():
            # Send double break if all is not well, and re-check
            self.updi_phy.send_double_break()
            self._init_session_parameters()
            if not self._check_datalink():
                raise PymcuprogSerialUpdiError("UPDI initialisation failed")

    def _check_datalink(self):
        """
        Check UPDI by loading CS STATUSA

        :returns: True if OK, False otherwise
        :rtype: bool
        """
        try:
            if self.ldcs(constants.UPDI_CS_STATUSA) != 0:
                self.logger.info("UPDI init OK")
                return True
        except PymcuprogSerialUpdiError:
            self.logger.info("UPDI datalink check failed")
            return False
        self.logger.info("UPDI not OK - reinitialisation required")
        return False

    def ldcs(self, address):
        """
        Load data from Control/Status space

        :param address: address to load
        :type address: int
        """
        self.logger.debug("LDCS from 0x%02X", address)
        self.updi_phy.send([constants.UPDI_PHY_SYNC, constants.UPDI_LDCS | (address & 0x0F)])
        response = self.updi_phy.receive(self.LDCS_RESPONSE_BYTES)
        numbytes_received = len(response)
        if numbytes_received != self.LDCS_RESPONSE_BYTES:
            raise PymcuprogSerialUpdiProtocolError("Unexpected number of bytes in response: "
                                 "{} byte(s), expected {} byte(s)".format(numbytes_received, self.LDCS_RESPONSE_BYTES))

        return response[0]

    def stcs(self, address, value):
        """
        Store a value to Control/Status space

        :param address: address to store to
        :type address: int
        :param value: value to write
        :type value: byte
        """
        self.logger.debug("STCS to 0x%02X", address)
        self.updi_phy.send([constants.UPDI_PHY_SYNC, constants.UPDI_STCS | (address & 0x0F), value])

    def ld_ptr_inc(self, size):
        """
        Loads a number of bytes from the pointer location with pointer post-increment

        :param size: number of bytes to load
        :type size: int
        :return: values read
        :rtype: list of bytes
        """
        self.logger.debug("LD8 from ptr++")
        self.updi_phy.send([constants.UPDI_PHY_SYNC, constants.UPDI_LD | constants.UPDI_PTR_INC |
                            constants.UPDI_DATA_8])
        return self.updi_phy.receive(size)

    def ld_ptr_inc16(self, words):
        """
        Load a 16-bit word value from the pointer location with pointer post-increment

        :param words: number of words to load
        :type words: int
        :return: values read
        :rtype: list of bytes
        """
        self.logger.debug("LD16 from ptr++")
        self.updi_phy.send([constants.UPDI_PHY_SYNC, constants.UPDI_LD | constants.UPDI_PTR_INC |
                            constants.UPDI_DATA_16])
        return self.updi_phy.receive(words << 1)

    def st_ptr_inc(self, data):
        """
        Store data to the pointer location with pointer post-increment

        :param data: data to store
        :type data: list of bytes
        """
        self.logger.debug("ST8 to *ptr++")
        self.updi_phy.send([constants.UPDI_PHY_SYNC, constants.UPDI_ST | constants.UPDI_PTR_INC | constants.UPDI_DATA_8,
                            data[0]])
        response = self.updi_phy.receive(1)

        if len(response) != 1 or response[0] != constants.UPDI_PHY_ACK:
            raise PymcuprogSerialUpdiProtocolError("ACK error with st_ptr_inc")

        num = 1
        while num < len(data):
            self.updi_phy.send([data[num]])
            response = self.updi_phy.receive(1)

            if len(response) != 1 or response[0] != constants.UPDI_PHY_ACK:
                raise PymcuprogSerialUpdiProtocolError("Error with st_ptr_inc")
            num += 1

    def st_ptr_inc16(self, data):
        """
        Store a 16-bit word value to the pointer location with pointer post-increment.
        ACK is disabled for blocks (> 2 bytes)

        :param data: data to store
        :type data: list of bytes
        """
        self.logger.debug("ST16 to *ptr++")
        if len(data) == 2:
            # ACKed mode used for 2-byte transfers
            self.logger.debug("ACKed block write")
            self.updi_phy.send([constants.UPDI_PHY_SYNC, constants.UPDI_ST | constants.UPDI_PTR_INC |
                                constants.UPDI_DATA_16, data[0], data[1]])
            response = self.updi_phy.receive(1)

            if len(response) != 1 or response[0] != constants.UPDI_PHY_ACK:
                raise PymcuprogSerialUpdiProtocolError("ACK error with st_ptr_inc16")
        else:
            self.logger.debug("ACKless block write")
            self._disable_ack()
            self.updi_phy.send([constants.UPDI_PHY_SYNC, constants.UPDI_ST | constants.UPDI_PTR_INC |
                            constants.UPDI_DATA_16])
            # Send data with no ACK (Response Signature)
            self.updi_phy.send(data)
            self._enable_ack()

    def repeat(self, repeats):
        """
        Store a value to the repeat counter

        :param repeats: number of repeats requested
        :type repeats: int
        """
        self.logger.debug("Repeat %d", repeats)
        if (repeats - 1) > constants.UPDI_MAX_REPEAT_SIZE:
            self.logger.error("Invalid repeat count of %d", repeats)
            raise PymcuprogSerialUpdiProtocolError("Invalid repeat count!")
        repeats -= 1
        self.updi_phy.send([constants.UPDI_PHY_SYNC, constants.UPDI_REPEAT | constants.UPDI_REPEAT_BYTE,
                            repeats & 0xFF])

    def read_sib(self):
        """
        Read the SIB

        :returns: SIB string
        :rtype: bytearray
        """
        return self.updi_phy.sib()

    def key(self, size, key):
        """
        Write a key

        :param size: size of key (0=64B, 1=128B, 2=256B)
        :type size: int
        :param key: key value
        :type key: list of bytes
        """
        self.logger.debug("Writing key")
        if len(key) != 8 << size:
            raise PymcuprogSerialUpdiProtocolError("Invalid KEY length!")
        self.updi_phy.send([constants.UPDI_PHY_SYNC, constants.UPDI_KEY | constants.UPDI_KEY_KEY | size])
        self.updi_phy.send(list(reversed(list(key))))

    def _st_data_phase(self, values):
        """
        Performs data phase of transaction:
        * receive ACK
        * send data

        :param values: value(s) to send
        :type values: list of bytes
        """
        response = self.updi_phy.receive(1)
        if len(response) != 1 or response[0] != constants.UPDI_PHY_ACK:
            raise PymcuprogSerialUpdiProtocolError("Error with st")

        self.updi_phy.send(values)
        response = self.updi_phy.receive(1)
        if len(response) != 1 or response[0] != constants.UPDI_PHY_ACK:
            raise PymcuprogSerialUpdiProtocolError("Error with st")


class UpdiDatalink16bit(UpdiDatalink):
    """
    UPDI data link layer in 16-bit version.
    This means that all addresses and pointers contain 2 bytes.
    """

    def __init__(self):
        UpdiDatalink.__init__(self)
        self.logger = getLogger(__name__)

    # pylint: disable=invalid-name
    def ld(self, address):
        """
        Load a single byte direct from a 16-bit address

        :param address: address to load from
        :type address: int
        :return: value read
        :rtype: bytes
        """
        self.logger.debug("LD from 0x%06X", address)
        self.updi_phy.send(
            [constants.UPDI_PHY_SYNC, constants.UPDI_LDS | constants.UPDI_ADDRESS_16 | constants.UPDI_DATA_8,
             address & 0xFF, (address >> 8) & 0xFF])
        return self.updi_phy.receive(1)[0]

    def ld16(self, address):
        """
        Load a 16-bit word directly from a 16-bit address

        :param address: address to load from
        :type address: int
        :return: values read
        :rtype: list of bytes
        """
        self.logger.debug("LD from 0x%06X", address)
        self.updi_phy.send(
            [constants.UPDI_PHY_SYNC, constants.UPDI_LDS | constants.UPDI_ADDRESS_16 | constants.UPDI_DATA_16,
             address & 0xFF, (address >> 8) & 0xFF])
        return self.updi_phy.receive(2)

    # pylint: disable=invalid-name
    def st(self, address, value):
        """
        Store a single byte value directly to a 16-bit address

        :param address: address to write to
        :type address: int
        :param value: value to write
        :type value: byte
        """
        self.logger.debug("ST to 0x%06X", address)
        self.updi_phy.send(
            [constants.UPDI_PHY_SYNC, constants.UPDI_STS | constants.UPDI_ADDRESS_16 | constants.UPDI_DATA_8,
             address & 0xFF, (address >> 8) & 0xFF])
        return self._st_data_phase([value & 0xFF])

    def st16(self, address, value):
        """
        Store a 16-bit word value directly to a 16-bit address

        :param address: address to write to
        :type address: int
        :param value: value to write
        :type value: byte
        """
        self.logger.debug("ST to 0x%06X", address)
        self.updi_phy.send(
            [constants.UPDI_PHY_SYNC, constants.UPDI_STS | constants.UPDI_ADDRESS_16 | constants.UPDI_DATA_16,
             address & 0xFF, (address >> 8) & 0xFF])
        return self._st_data_phase([value & 0xFF, (value >> 8) & 0xFF])

    def st_ptr(self, address):
        """
        Set the pointer location

        :param address: address to write
        :type address: int
        """
        self.logger.debug("ST to ptr")
        self.updi_phy.send(
            [constants.UPDI_PHY_SYNC, constants.UPDI_ST | constants.UPDI_PTR_ADDRESS | constants.UPDI_DATA_16,
             address & 0xFF, (address >> 8) & 0xFF])
        response = self.updi_phy.receive(1)
        if len(response) != 1 or response[0] != constants.UPDI_PHY_ACK:
            raise PymcuprogSerialUpdiProtocolError("Error with st_ptr")


class UpdiDatalink24bit(UpdiDatalink):
    """
    UPDI data link layer in 24-bit version.
    This means that all addresses and pointers contain 3 bytes
    """

    def __init__(self):
        UpdiDatalink.__init__(self)
        self.logger = getLogger(__name__)

    # pylint: disable=invalid-name
    def ld(self, address):
        """
        Load a single byte direct from a 24-bit address

        :param address: address to load from
        :type address: int
        :return: value read
        :rtype: bytes
        """
        self.logger.debug("LD from 0x%06X", address)
        self.updi_phy.send(
            [constants.UPDI_PHY_SYNC, constants.UPDI_LDS | constants.UPDI_ADDRESS_24 | constants.UPDI_DATA_8,
             address & 0xFF, (address >> 8) & 0xFF, (address >> 16) & 0xFF])
        return self.updi_phy.receive(1)[0]

    def ld16(self, address):
        """
        Load a 16-bit word directly from a 24-bit address

        :param address: address to load from
        :type address: int
        :return: values read
        :rtype: list of bytes
        """
        self.logger.debug("LD from 0x%06X", address)
        self.updi_phy.send(
            [constants.UPDI_PHY_SYNC, constants.UPDI_LDS | constants.UPDI_ADDRESS_24 | constants.UPDI_DATA_16,
             address & 0xFF, (address >> 8) & 0xFF, (address >> 16) & 0xFF])
        return self.updi_phy.receive(2)

    # pylint: disable=invalid-name
    def st(self, address, value):
        """
        Store a single byte value directly to a 24-bit address

        :param address: address to write to
        :type address: int
        :param value: value to write
        :type value: byte
        """
        self.logger.debug("ST to 0x%06X", address)
        self.updi_phy.send(
            [constants.UPDI_PHY_SYNC, constants.UPDI_STS | constants.UPDI_ADDRESS_24 | constants.UPDI_DATA_8,
             address & 0xFF, (address >> 8) & 0xFF, (address >> 16) & 0xFF])
        return self._st_data_phase([value & 0xFF])

    def st16(self, address, value):
        """
        Store a 16-bit word value directly to a 24-bit address

        :param address: address to write to
        :type address: int
        :param value: value to write
        :type value: byte
        """
        self.logger.debug("ST to 0x%06X", address)
        self.updi_phy.send(
            [constants.UPDI_PHY_SYNC, constants.UPDI_STS | constants.UPDI_ADDRESS_24 | constants.UPDI_DATA_16,
             address & 0xFF, (address >> 8) & 0xFF, (address >> 16) & 0xFF])
        return self._st_data_phase([value & 0xFF, (value >> 8) & 0xFF])

    def st_ptr(self, address):
        """
        Set the pointer location

        :param address: address to write
        :type address: int
        """
        self.logger.debug("ST to ptr")
        self.updi_phy.send(
            [constants.UPDI_PHY_SYNC, constants.UPDI_ST | constants.UPDI_PTR_ADDRESS | constants.UPDI_DATA_24,
             address & 0xFF, (address >> 8) & 0xFF, (address >> 16) & 0xFF])
        response = self.updi_phy.receive(1)
        if len(response) != 1 or response[0] != constants.UPDI_PHY_ACK:
            raise PymcuprogSerialUpdiProtocolError("Error with st_ptr")
